import gc
import sys
import threading

from django.template import RequestContext
from django.shortcuts import render
from django.utils.html import escape
from types import ModuleType, FrameType
try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from django.contrib.auth.decorators import user_passes_test

from . import dowser, DOWSER_NAMES
from . import reftree


@user_passes_test(lambda u: u.is_superuser)
def index(request):
    floor = int(request.GET.get('floor', default=0))
    rows = []
    typenames = sorted(dowser.history.keys())
    for typename in typenames:
        history = dowser.history[typename]
        maxhist = 0
        for hist in history:
            maxhist = max(maxhist, max(hist))
        charts = " ".join(['<img class="chart" src="%s" alt="%s"/>' %
                           (chart_url2(history[x]), DOWSER_NAMES[x])
                           for x in range(len(history))])
        if maxhist > floor:
            row = ('<div class="typecount">%s<br />'
                   '%s<br />'
                   'Cur: %s Max: %s <a href="%s">TRACE</a></div>'
                   % (escape(typename),
                      charts,
                      history[0][0], maxhist,
                      "trace/%s" % typename,
                      )
                   )
            rows.append(row)
    return render(request, "django_dowser/graphs.html",
                  {'output': "\n".join(rows)})


def chart_url2(entries):
    url = ("http://chart.apis.google.com/chart?chs=%sx20&cht=ls&"
           "chco=0077CC&chd=t:%s") % (
               len(entries), ",".join(str(x) for x in reversed(entries)))
    return url


def chart_url(typename, history_slot=0):
    data = reversed(list(dowser.history[typename][history_slot]))  # TODO
    url = chart_url2(data)
    return url


@user_passes_test(lambda u: u.is_superuser)
def trace(request, typename, objid=None):
    gc.collect()

    if objid is None:
        rows = trace_all(typename)
    else:
        rows = trace_one(typename, objid)

    return render(request, "django_dowser/trace.html", {
        'output': "\n".join(rows),
        'typename': typename,
        'objid': objid or "",
    })


@user_passes_test(lambda u: u.is_superuser)
def tree(request, objid, typename):
    gc.collect()

    rows = []
    objid = int(objid)
    all_objs = gc.get_objects()
    for obj in all_objs:
        if id(obj) == objid:
            objtype = type(obj)
            if str(objtype.__module__) + "." + objtype.__name__ != typename:
                rows = ["<h3>The object you requested is no longer "
                        "of the correct type.</h3>"]
            else:
                rows.append('<div class="obj">')

                my_tree = ReferrerTree(obj)
                my_tree.ignore(all_objs)
                for depth, parentid, parentrepr in my_tree.walk(
                        maxresults=1000):
                    rows.append(parentrepr)

                rows.append('</div>')
            break
    if not rows:
        rows = ["<h3>The object you requested was not found.</h3>"]

    return render(request, "django_dowser/tree.html", {
        'output': "\n".join(rows), 'typename': typename, 'objid': objid})


method_types = [
    type(tuple.__le__),                 # 'wrapper_descriptor'
    type([1].__le__),                   # 'method-wrapper'
    type(sys.getcheckinterval),         # 'builtin_function_or_method'
    type(threading.Thread.isAlive),     # 'instancemethod'
]


def trace_all(typename):
    rows = []
    for obj in gc.get_objects():
        objtype = type(obj)
        if str(objtype.__module__) + "." + objtype.__name__ == typename:
            rows.append("<p class='obj'>%s</p>"
                        % ReferrerTree(obj).get_repr(obj))
    if not rows:
        rows = ["<h3>The type you requested was not found.</h3>"]
    return rows


def trace_one(typename, objid):
    rows = []
    objid = int(objid)
    all_objs = gc.get_objects()
    for obj in all_objs:
        if id(obj) == objid:
            objtype = type(obj)
            if str(objtype.__module__) + "." + objtype.__name__ != typename:
                rows = ["<h3>The object you requested is no longer "
                        "of the correct type.</h3>"]
            else:
                tree = ReferrerTree(obj)

                # repr
                rows.append("<p class='obj'>%s</p>" % get_repr(obj, 5000))

                # Attributes
                rows.append('<div class="obj"><h3>Attributes</h3>')
                for k in dir(obj):
                    v = getattr(obj, k)
                    if type(v) not in method_types:
                        rows.append('<p class="attr"><b>%s:</b> %s</p>' %
                                    (k, get_repr(v)))
                    del v
                rows.append('</div>')

                # Referrers
                rows.append('<div class="refs"><h3>Referrers (Parents)</h3>')
                rows.append('<p class="desc"><a href="%s">Show the '
                            'entire tree</a> of reachable objects</p>'
                            % ("../../tree/%s/%s" % (typename, objid)))
                tree.ignore(all_objs)
                for depth, parentid, parentrepr in tree.walk(maxdepth=1):
                    if parentid:
                        rows.append("<p class='obj'>%s</p>" % parentrepr)
                rows.append('</div>')

                # Referents
                rows.append('<div class="refs"><h3>Referents (Children)</h3>')
                for child in gc.get_referents(obj):
                    rows.append("<p class='obj'>%s</p>" % tree.get_repr(child))
                rows.append('</div>')
            break
    if not rows:
        rows = ["<h3>The object you requested was not found.</h3>"]
    return rows


def get_repr(obj, limit=250):
    return escape(reftree.get_repr(obj, limit))


class ReferrerTree(reftree.Tree):

    ignore_modules = True

    def _gen(self, obj, depth=0):
        if self.maxdepth and depth >= self.maxdepth:
            yield depth, 0, "---- Max depth reached ----"
            return
        if isinstance(obj, ModuleType) and self.ignore_modules:
            return

        refs = gc.get_referrers(obj)
        refiter = iter(refs)
        self.ignore(refs, refiter)
        thisfile = sys._getframe().f_code.co_filename
        for ref in refiter:
            # Exclude all frames that are from this module or reftree.
            if isinstance(ref, FrameType) and ref.f_code.co_filename in (
                    thisfile, self.filename):
                continue

            # Exclude all functions and classes from this module or reftree.
            mod = getattr(ref, "__module__", "")
            if "django_dowser" in mod or "reftree" in mod or mod == '__main__':
                continue

            # Exclude all parents in our ignore list.
            if id(ref) in self._ignore:
                continue

            # Yield the (depth, id, repr) of our object.
            yield depth, 0, '%s<div class="branch">' % (" " * depth)
            if id(ref) in self.seen:
                yield depth, id(ref), "see %s above" % id(ref)
            else:
                self.seen[id(ref)] = None
                yield depth, id(ref), self.get_repr(ref, obj)

                for parent in self._gen(ref, depth + 1):
                    yield parent
            yield depth, 0, '%s</div>' % (" " * depth)

    def get_repr(self, obj, referent=None):
        """Return an HTML tree block describing the given object."""
        objtype = type(obj)
        typename = str(objtype.__module__) + "." + objtype.__name__
        prettytype = typename.replace("__builtin__.", "")

        name = getattr(obj, "__name__", "")
        if name:
            prettytype = "%s %r" % (prettytype, name)

        key = ""
        if referent:
            key = self.get_refkey(obj, referent)
        url = reverse('dowser_trace_object', args=(
            typename,
            id(obj)
        ))
        return ('<a class="objectid" href="%s">%s</a> '
                '<span class="typename">%s</span>%s<br />'
                '<span class="repr">%s</span>'
                % (url, id(obj), prettytype, key, get_repr(obj, 100))
                )

    def get_refkey(self, obj, referent):
        """Return the dict key or attribute name of obj which refers to
        referent."""
        if isinstance(obj, dict):
            for k, v in obj.items():
                if v is referent:
                    return " (via its %r key)" % k

        for k in dir(obj) + ['__dict__']:
            if getattr(obj, k, None) is referent:
                return " (via its %r attribute)" % k
        return ""



try:
    from django.conf.urls import url
except ImportError:
    from django.conf.urls.defaults import url

from . import views

urlpatterns = [
    url(r'^trace/(?P<typename>[\.\-\w]+)$', views.trace, name='dowser_trace_type'),
    url(r'^trace/(?P<typename>[\.\-\w]+)/(?P<objid>\d+)$', views.trace, name='dowser_trace_object'),
    url(r'^tree/(?P<typename>[\.\-\w]+)/(?P<objid>\d+)$', views.tree, name='dowser_tree'),
    url(r'^$', views.index, name='dowser_index'),
]



import os
from setuptools import setup

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

VERSION = '0.1.7'

setup(
    name="django-dowser",
    version=VERSION,

    author="Mateusz Lapsa-Malawski",
    author_email="m@cr3.io",
    url="http://github.com/munhitsu/django-dowser",

    description="Django fork of amazing memory leaks tracking application for python wsgi - the Dowser",
    long_description=open("README.rst").read(),
    keywords=["django", "dowser", "debug", "memory leak"],
    license=open("LICENSE").read(),
    download_url="http://pypi.python.org/packages/source/d/django-dowser/django-dowser-%s.tar.gz" % VERSION,
    platforms="",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Environment :: Other Environment",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: BSD License",
    ],
    packages=['django_dowser'],
    include_package_data=True,
    zip_safe=False,  # because we're including media that Django needs
)



import gc
import sys
import logging
from types import FrameType

log = logging.getLogger(__name__)


class Tree(object):
    def __init__(self, obj):
        self.seen = {}
        self.maxdepth = None
        self.obj = obj
        self.filename = sys._getframe().f_code.co_filename
        self._ignore = {}

    def ignore(self, *objects):
        for obj in objects:
            self._ignore[id(obj)] = None

    def ignore_caller(self):
        f = sys._getframe()     # = this function
        cur = f.f_back          # = the function that called us (probably 'walk')
        self.ignore(cur, cur.f_builtins, cur.f_locals, cur.f_globals)
        caller = f.f_back       # = the 'real' caller
        self.ignore(caller, caller.f_builtins, caller.f_locals, caller.f_globals)

    def walk(self, maxresults=100, maxdepth=None):
        """Walk the object tree, ignoring duplicates and circular refs."""
        log.debug("step")
        self.seen = {}
        self.ignore(self, self.__dict__, self.obj, self.seen, self._ignore)

        # Ignore the calling frame, its builtins, globals and locals
        self.ignore_caller()
        self.maxdepth = maxdepth
        count = 0
        log.debug("will iterate results")
        for result in self._gen(self.obj):
            log.debug("will yeld")
            yield result
            count += 1
            if maxresults and count >= maxresults:
                yield 0, 0, "==== Max results reached ===="
                return

    def print_tree(self, maxresults=100, maxdepth=None):
        """Walk the object tree, pretty-printing each branch."""
        self.ignore_caller()
        for depth, refid, rep in self.walk(maxresults, maxdepth):
            print(("%9d" % refid), (" " * depth * 2), rep)


def _repr_container(obj):
    return "%s of len %s: %r" % (type(obj).__name__, len(obj), obj)
repr_dict = _repr_container
repr_set = _repr_container
repr_list = _repr_container
repr_tuple = _repr_container


def repr_str(obj):
    return "%s of len %s: %r" % (type(obj).__name__, len(obj), obj)
repr_unicode = repr_str


def repr_frame(obj):
    return "frame from %s line %s" % (obj.f_code.co_filename, obj.f_lineno)


def get_repr(obj, limit=250):
    typename = getattr(type(obj), "__name__", None)
    handler = globals().get("repr_%s" % typename, repr)

    try:
        result = handler(obj)
    except:
        result = "unrepresentable object: %r" % sys.exc_info()[1]

    if len(result) > limit:
        result = result[:limit] + "..."

    return result


class ReferentTree(Tree):
    def _gen(self, obj, depth=0):
        if self.maxdepth and depth >= self.maxdepth:
            yield depth, 0, "---- Max depth reached ----"
            return

        for ref in gc.get_referents(obj):
            if id(ref) in self._ignore:
                continue
            elif id(ref) in self.seen:
                yield depth, id(ref), "!" + get_repr(ref)
                continue
            else:
                self.seen[id(ref)] = None
                yield depth, id(ref), get_repr(ref)
            for child in self._gen(ref, depth + 1):
                yield child


class ReferrerTree(Tree):
    def _gen(self, obj, depth=0):
        if self.maxdepth and depth >= self.maxdepth:
            yield depth, 0, "---- Max depth reached ----"
            return

        refs = gc.get_referrers(obj)
        refiter = iter(refs)
        self.ignore(refs, refiter)
        for ref in refiter:
            # Exclude all frames that are from this module.
            if isinstance(ref, FrameType):
                if ref.f_code.co_filename == self.filename:
                    continue
            if id(ref) in self._ignore:
                continue
            elif id(ref) in self.seen:
                yield depth, id(ref), "!" + get_repr(ref)
                continue
            else:
                self.seen[id(ref)] = None
                yield depth, id(ref), get_repr(ref)
            for parent in self._gen(ref, depth + 1):
                yield parent


class CircularReferents(Tree):
    def walk(self, maxresults=100, maxdepth=None):
        """Walk the object tree, showing circular referents."""
        self.stops = 0
        self.seen = {}
        self.ignore(self, self.__dict__, self.seen, self._ignore)

        # Ignore the calling frame, its builtins, globals and locals
        self.ignore_caller()

        self.maxdepth = maxdepth
        count = 0
        for result in self._gen(self.obj):
            yield result
            count += 1
            if maxresults and count >= maxresults:
                yield 0, 0, "==== Max results reached ===="
                return

    def _gen(self, obj, depth=0, trail=None):
        if self.maxdepth and depth >= self.maxdepth:
            self.stops += 1
            return

        if trail is None:
            trail = []

        for ref in gc.get_referents(obj):
            if id(ref) in self._ignore:
                continue
            elif id(ref) in self.seen:
                continue
            else:
                self.seen[id(ref)] = None

            refrepr = get_repr(ref)
            if id(ref) == id(self.obj):
                yield trail + [refrepr]

            for child in self._gen(ref, depth + 1, trail + [refrepr]):
                yield child

    def print_tree(self, maxresults=100, maxdepth=None):
        """Walk the object tree, pretty-printing each branch."""
        self.ignore_caller()
        for trail in self.walk(maxresults, maxdepth):
            print(trail)
        if self.stops:
            print("%s paths stopped because max depth reached" % self.stops)


def count_objects():
    d = {}
    for obj in gc.get_objects():
        objtype = type(obj)
        d[objtype] = d.get(objtype, 0) + 1
    d = sorted((v, k) for k, v in d.items())
    return d



import gc
import threading
import time
from collections import deque
import itertools

# maximum from all old periods is being promoted to next one
DOWSER_MAXENTRIES = [12, 120, 60, 60]
DOWSER_TICKS = [5, 6, 48, 28]
DOWSER_NAMES = ["1m", "1h", "1d", "4w"]


class Dowser(object):
    """
    A working thread to gather type usage
    """
    history = {}
    samples = []

    def __init__(self):
        # TODO: how to limit it only to server process not the monitor
        # TODO: cover multi-process configuration - maybe as separate daemon...
        self.running = False
        self.samples = [0] * len(DOWSER_MAXENTRIES)
        self.runthread = threading.Thread(target=self.start)
        self.runthread.daemon = True
        self.runthread.start()

    def start(self):
        self.running = True
        while self.running:
            self.tick()
            time.sleep(DOWSER_TICKS[0])

    def tick(self):
        gc.collect()
        typecounts = {}

        for obj in gc.get_objects():
            objtype = type(obj)
            typename = str(objtype.__module__) + "." + objtype.__name__
            if typename in typecounts:
                typecounts[typename] += 1
            else:
                typecounts[typename] = 1

        for typename, count in typecounts.items():
            if typename not in self.history:
                self.history[typename] = [deque([0] * x) for x in DOWSER_MAXENTRIES]
            self.history[typename][0].appendleft(count)

        self.samples[0] += 1
        promote = [False] * (len(DOWSER_MAXENTRIES)-1)

        # let's calculate what we promote
        for i in range(len(self.samples)-1):
            if self.samples[i] >= DOWSER_TICKS[i]:
                promote[i] = True
                self.samples[i+1] += 1
                self.samples[i] = 0

        for typename, hist in self.history.items():
            history = self.history[typename]
            # let's promote max from (set of entries to lower granularity history)
            for i in range(len(self.samples)-1):
                if promote[i]:
                    history[i+1].appendleft(max(itertools.islice(history[i], 0, DOWSER_TICKS[i])))
            # let's limit history to DOWSER_MAXENTRIES
            for i in range(len(self.samples)):
                if len(history[i]) > DOWSER_MAXENTRIES[i]:
                    history[i].pop()

    def stop(self):
        self.running = False


dowser = Dowser()
